import argparse
import datetime
import logging
import os
import socketserver
import string
import struct
import threading
from lxml import etree
try:
    from cdrlib import db, search, xslt, xsd
except Exception as e:
    fp = open("c:/cdr/log/CdrServer.err", "a")
    fp.write("import failed: %s" % e)
    fp.close()
    raise

class ThreadHandler(socketserver.BaseRequestHandler):

    MAX_REQUEST_LENGTH = 150000000

    def handle(self):
        self.logger = logging.getLogger("cdr-server")
        self.start = datetime.datetime.now()
        self.session = None
        self.response = etree.Element("CdrResponseSet")
        self.response.set("Time", self.start.strftime("%Y-%m-%dT%H:%M:%S.%f"))
        length_bytes = self.request.recv(4)
        if len(length_bytes) != 4:
            self.logger.warning("request length not sent")
            return None
        length, = struct.unpack("!L", length_bytes)
        if not length or length > self.MAX_REQUEST_LENGTH:
            self.logger.warning("refusing %s-byte request", length)
            return None
        request_bytes = self.request.recv(4096)
        #print(str(request_bytes))
        if b"<CdrCommandSet" not in request_bytes:
            self.logger.warning("refusing bogus request string: %r",
                                request_bytes)
            return None
        received = len(request_bytes)
        chunks = [request_bytes]
        while received < length:
            request_bytes = self.request.recv(4096)
            if not request_bytes:
                break
            chunks.append(request_bytes)
            received += len(request_bytes)
        if received < length:
            error = ("Expected {} bytes reading client request; "
                     "got {} bytes").format(length, received)
            self.logger.warning(error)
            self.response.append(self.errors(error))
        else:
            self.logger.debug("received %d bytes", received)
            root = etree.fromstring(b"".join(chunks))
            if root.tag != "CdrCommandSet":
                error = self.errors("Top element must be CdrCommandSet")
                self.response.append(error)
            else:
                for node in root:
                    if node.tag == "SessionId":
                        self.session = node.text
                        self.logger.debug("session: %r", self.session)
                    if node.tag == "CdrCommand":
                        self.process_command(node)
        response = etree.tostring(self.response, encoding="utf-8")
        length_bytes = struct.pack("!L", len(response))
        self.request.send(length_bytes)
        self.request.sendall(response)
        #self.request.sendall(etree.tostring(self.response, pretty_print=True))
    def process_command(self, node):
        self.logger.debug("processing command")
        start = datetime.datetime.now()
        response = etree.Element("CdrResponse")
        response.set("Status", "success")
        cmd_id = node.get("CmdId")
        if cmd_id:
            self.logger.debug("CmdId: %r", cmd_id)
            response.set("CmdId", cmd_id)
        children = node.getchildren()
        if not children:
            self.logger.error("no child nodes for command")
            response.set("Status", "failure")
            response.append(self.errors("Missing specific command element"))
        elif len(children) > 1:
            response.set("Status", "failure")
            response.append(self.errors("Too many specific command elements"))
            self.logger.error("%d child nodes for command", len(children))
        else:
            try:
                child = children[0]
                #print("handling command", child.tag)
                self.logger.info(child.tag)
                if child.tag == "CdrListDocTypes":
                    response.append(self.list_doc_types())
                elif child.tag == "CdrCacheing":
                    response.append(self.control_cache(child))
                elif child.tag == "CdrFilter":
                    xslt = xslt.XSLT(child)
                    response.append(xslt.filter_doc())
                elif child.tag == "CdrValidateDoc":
                    validator = xsd.Validator(request=child)
                    response.append(validator.run())
                elif child.tag == "CdrSearch":
                    query = search.Query(request=child)
                    response.append(query.run())
                else:
                    self.logger.error("unknown command %r", child.tag)
                    raise Exception("Unknown command: %s" % child.tag)
            except Exception as e:
                self.logger.exception("process_command()")
                response.set("Status", "failure")
                message = "Unknown command: %s" % child.tag
                response.append(self.errors(str(e)))
        delta = (datetime.datetime.now() - start).total_seconds()
        response.set("Elapsed", "%f" % delta)
        self.response.append(response)
    def control_cache(self, node):
        term_cache_aliases = ("term", "pub", "all")
        if node.get("on") in term_cache_aliases:
            on = True
        elif node.get("off") in term_cache_aliases:
            on = False
        else:
            raise Exception("No legal value found for on/off attribute"
                            " in CdrCacheing command")
        xslt.Term.set_cache(on)
        return etree.Element("CdrCacheingResp")
    def list_doc_types(self):
        query = db.Query("doc_type", "name").order("name")
        query.where("active = 'Y'")
        response = etree.Element("CdrListDocTypesResp")
        for row in query.execute().fetchall():
            etree.SubElement(response, "DocType").text = row[0]
        return response
    def errors(self, *messages):
        errors = etree.Element("Errors")
        for message in messages:
            etree.SubElement(errors, "Err").text = message
        return errors

class Local(threading.local):
    def __init__(self, **kw):
        self.conn = db.connect()
        self.__dict__.update(kw)

class Server(socketserver.ThreadingMixIn, socketserver.TCPServer):
    local = Local()
    WORKING_DRIVE = None
    HOST = "localhost"
    PORT = 2019
    request_queue_size = 500
    allow_reuse_address = True
    #def server_activate(self):
    #    self.socket.listen(256)
    @classmethod
    def get_working_drive(cls):
        if cls.WORKING_DRIVE is None:
            for drive in string.ascii_uppercase[2:]:
                try:
                    os.stat(drive + ":/cdr/lib/Python")
                    cls.WORKING_DRIVE = drive
                    break
                except:
                    pass
        if cls.WORKING_DRIVE is None:
            raise Exception("unable to find CDR working files")
        return cls.WORKING_DRIVE
    @classmethod
    def run(cls):
        parser = argparse.ArgumentParser()
        parser.add_argument("--log-level", default="info")
        args = parser.parse_args()
        logger = Logging.get_logger("cdr-server", level=args.log_level)
        server = cls((cls.HOST, cls.PORT), ThreadHandler)
        msg = "listening at %s on port %s" % server.server_address
        print(msg)
        logger.info(msg)
        #server_thread = threading.Thread(target=server.serve_forever)
        #server_thread.daemon = True
        #server_thread.start()
        #print("Server loop running in thread", server_thread.name)
        server.serve_forever()

class Logging:
    FORMAT = "%(asctime)s [%(levelname)s] %(message)s"
    LEVELS = {
        "info": logging.INFO,
        "debug": logging.DEBUG,
        "warn": logging.WARN,
        "warning": logging.WARNING,
        "critical": logging.CRITICAL,
        "error": logging.ERROR
    }
    class Formatter(logging.Formatter):
        """Make our own logging formatter to get the time stamps right."""
        converter = datetime.datetime.fromtimestamp
        def formatTime(self, record, datefmt=None):
            ct = self.converter(record.created)
            if datefmt:
                s = ct.strftime(datefmt)
            else:
                t = ct.strftime("%Y-%m-%d %H:%M:%S")
                s = "%s.%03d" % (t, record.msecs)
            return s
    @classmethod
    def get_logger(cls, name, **opts):
        """
        Factory method for instantiating a logging object.

        name       required name for the logger
        path       optional path for the log file
        format     optional override for the default log format pattern
        level      optional verbosity for logging, defaults to info
        propagate  if True, the base handler also writes our entries
        multiplex  if True, add new handler even there already is one
        """

        logger = logging.getLogger(name)
        level = opts.get("level", "info")
        logger.setLevel(cls.LEVELS.get(level, logging.INFO))
        logger.propagate = opts.get("propagate", False)
        if not logger.handlers or opts.get("multiplex"):
            working_drive = Server.get_working_drive()
            default_path = "%s:/cdr/Log/%s.log" % (working_drive, name)
            path = opts.get("path", default_path)
            handler = logging.FileHandler(path)
            formatter = cls.Formatter(opts.get("format", cls.FORMAT))
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        return logger

Server.run()
