import datetime
import requests
import socket
import struct
import time
from lxml import etree
from cdrlib import util

class API:
    logger = util.Logging.get_logger("cdr-client")
    class Request:
        DEFAULT_HOST = "127.0.0.1"
        DEFAULT_PORT = 2019
        DEFAULT_TIMEOUT = 5 * 60
        DEFAULT_SLEEP = 3
        def __init__(self, *commands, **opts):
            session = self.Session(**opts)
            self.root = etree.Element("CdrCommandSet")
            etree.SubElement(self.root, "SessionId").text = str(session)
            for command in commands:
                cid = str(time.time())
                wrapper = etree.SubElement(self.root, "CdrCommand", CmdId=cid)
                wrapper.append(command)
        def check_time(self):
            now = datetime.datetime.now()
            if now > self.quitting_time:
                error = "timeout (%f seconds) exceeded" % self.timeout
                API.logger.error(error)
                raise Exception(error)
            remaining = (self.quitting_time - now).total_seconds()
            time.sleep(min(self.sleep, remaining))
        def submit(self, **opts):
            start = datetime.datetime.now()
            host = opts.get("host") or self.DEFAULT_HOST
            port = int(opts.get("port") or self.DEFAULT_PORT)
            self.sleep = float(opts.get("sleep") or self.DEFAULT_SLEEP)
            self.timeout = float(opts.get("timeout") or self.DEFAULT_TIMEOUT)
            timedelta = datetime.timedelta(seconds=self.timeout)
            self.quitting_time = start + timedelta
            command_set = etree.tostring(self.root, encoding="utf-8")
            API.logger.debug(command_set)
            length = len(command_set)
            length_bytes = struct.pack("!L", length)
            while True:
                try:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    API.logger.debug("connect(%s:%s)", host, port)
                    sock.connect((host, port))
                except Exception as e:
                    API.logger.error("sock.connect(): %s", e)
                    self.check_time()
                    sock.close()
                    continue
                try:
                    sock.send(length_bytes)
                    sock.sendall(command_set)
                    length_bytes = sock.recv(4)
                    if len(length_bytes) != 4:
                        raise Exception("server did not send response length")
                    length, = struct.unpack("!L", length_bytes)
                    if not length:
                        raise Exception("server did not send response")
                    received = 0
                    chunks = []
                    while received < length:
                        response_bytes = sock.recv(8192)
                        if not response_bytes:
                            break
                        chunks.append(response_bytes)
                        received += len(response_bytes)
                    if received < length:
                        pattern = "Expected {} bytes, got {} bytes"
                        raise Exception(pattern.format(length, received))
                    response = b"".join(chunks)
                    API.logger.debug(response)
                    return self.ResponseSet(response)
                except Exception as e:
                    API.logger.error("Request.submit(): %s", e)
                    self.check_time()
                    continue
                finally:
                    sock.close()
        class Session:
            DEFAULT = "guest"
            def __init__(self, **opts):
                self.user = opts.get("user")
                if self.user:
                    self.name = API.login(self.user, opts.get("password"))
                else:
                    self.name = opts.get("session") or self.DEFAULT
            def __str__(self):
                return self.name
        class ResponseSet:
            def __init__(self, server_bytes):
                self.bytes = server_bytes
                self.root = etree.fromstring(server_bytes)
                self.time = self.root.get("Time")
                self.responses = []
                self.errors = []
                for node in self.root:
                    if node.tag == "Errors":
                        for err in node.findall("Err"):
                            self.errors.append(err.text)
                    elif node.tag == "CdrResponse":
                        self.responses.append(self.Response(node))
            class Response:
                def __init__(self, node):
                    self.tag = node.tag
                    self.command_id = node.get("CmdId")
                    children = node.getchildren()
                    message = "exactly one response child node required"
                    assert len(children) == 1, message
                    self.payload = children[0]
    @classmethod
    def login(cls, username, password):
        url = "https://cdr-dev.cancer.gov/cgi-bin/secure/login.py"
        requests.packages.urllib3.disable_warnings()
        auth = requests.auth.HTTPDigestAuth(username, password)
        response = requests.get(url, auth=auth, verify=False)
        session = response.text.strip()
        cls.logger.debug("new session %r", session)
        return session
    @classmethod
    def search(cls, **opts):
        class Result:
            def __init__(self, node):
                self.id = util.Doc.get_text(node.find("DocId"))
                self.type = util.Doc.get_text(node.find("DocType"))
                self.title = util.Doc.get_text(node.find("DocTitle"))
            def __str__(self):
                return "{} ({}) {}".format(self.id, self.type, self.title)
        session = opts.get("session")
        tests = opts.get("tests") or []
        limit = opts.get("limit")
        timeout = opts.get("timeout")
        doc_types = opts.get("doc_types") or []
        command = etree.Element("CdrSearch")
        query = etree.SubElement(command, "Query")
        if limit:
            query.set("MaxDocs", str(limit))
        if isinstance(tests, str):
            tests = [tests]
        if isinstance(doc_types, str):
            doc_types = [doc_types]
        for doc_type in doc_types:
            etree.SubElement(query, "DocType").text = doc_type
        for test in tests:
            etree.SubElement(query, "Test").text = test
        request = API.Request(command, session=session)
        response_set = request.submit(timeout=timeout, host=opts.get("host"))
        for response in response_set.responses:
            nodes = response.payload.findall("QueryResults/QueryResult")
            return [Result(node) for node in nodes]
        raise Exception("CdrSearchResponse node not received")
    class Command:
        def __init__(self, command):
            self.node = etree.Element("CdrCommand")
            self.node.append(command)

if __name__ == "__main__":
    from argparse import ArgumentParser
    from getpass import getpass
    API.logger.setLevel("DEBUG")
    default_test = "CdrCtl/Title contains breast reconstruction"
    parser = ArgumentParser()
    parser.add_argument("--test", nargs="*", default=default_test)
    parser.add_argument("--host")
    parser.add_argument("--port")
    parser.add_argument("--username")
    parser.add_argument("--session")
    parser.add_argument("--limit", type=int)
    parser.add_argument("--timeout", type=float, default=5.0)
    opts = parser.parse_args()
    name = opts.username
    session = opts.session or name and API.login(name, getpass()) or "guest"
    args = {
        "session": session,
        "tests": opts.test,
        "limit": opts.limit,
        "timeout": opts.timeout,
        "host": opts.host,
        "port": opts.port
        }
    for result in API.search(**args):
        print(result)
