import argparse
import datetime
import logging
import re
import sys
import unicodedata
from builtins import dict
from lxml import etree
from cdrlib import db, util, xslt

class Query:
    logger = util.Session.logger
    logger.setLevel("DEBUG")
    def __init__(self, **args):
        self.cursor = util.Session.local.cursor
        self.session = util.Session(args.get("session", "guest"))
        self.logger.debug("session=%r", self.session)
        self.doc_types = []
        self.tests = []
        self.max_docs = None
        if "request" in args:
            self.parse_request(args["request"])
        else:
            required = 'test must have "path operator value"'
            self.doc_types = args.get("doc_type") or []
            self.max_docs = args.get("max_docs")
            tests = args.get("test") or []
            if not isinstance(self.doc_types, list):
                self.doc_types = [self.doc_types]
            if not isinstance(tests, list):
                tests = [tests]
            self.tests = [self.Test(test) for test in tests]
    def parse_request(self, request):
        self.logger.debug("parsing request")
        query = request.find("Query")
        assert len(query), "Query node missing"
        self.max_docs = query.get("MaxDocs")
        self.logger.debug("max_docs: %r", self.max_docs)
        for node in query:
            if node.tag == "DocType":
                self.doc_types.append(node.text)
            elif node.tag == "Test":
                test = util.Doc.get_text(node)
                self.logger.debug("test %r", test)
                self.tests.append(self.Test(test))
    def run(self):
        docs = self.get_docs()
        count = str(len(docs))
        response = etree.Element("CdrSearchResp")
        results = etree.SubElement(response, "QueryResults", count=count)
        for doc in docs:
            results.append(doc.query_result())
        return response
    class Doc:
        def __init__(self, doc_id, doc_title, doc_type):
            self.doc_id = doc_id
            self.doc_title = doc_title
            self.doc_type = doc_type
        def query_result(self):
            result = etree.Element("QueryResult")
            etree.SubElement(result, "DocId").text = "CDR%010d" % self.doc_id
            etree.SubElement(result, "DocType").text = self.doc_type
            etree.SubElement(result, "DocTitle").text = self.doc_title
            return result
    def get_docs(self):
        rows = self.db_query().execute(util.Session.local.cursor).fetchall()
        return [self.Doc(*row) for row in rows]
    def db_query(self):
        fields = "d.id", "d.title", "t.name"
        dbq = db.Query("document d", *fields).order("d.title")
        dbq.join("doc_type t", "t.id = d.doc_type")
        if self.doc_types:
            if len(self.doc_types) == 1:
                dbq.where(dbq.Condition("t.name", self.doc_types[0]))
            else:
                dbq.where(dbq.Condition("t.name", self.doc_types, "IN"))
        n = 1
        for test in self.tests:
            if test.column:
                alias = "qt%s" % n
                join_test = "qt%d.doc_id"
                dbq.join("query_term qt%d" % n, "qt%d.doc_id = d.id" % n)
                dbq.where(dbq.Condition("qt%d.path" % n, test.path))
                column = "qt%d.%s" % (n, test.column)
                dbq.where(dbq.Condition(column, test.value, test.operator))
                n += 1
            else:
                dbq.where(dbq.Condition("d.title", test.value, test.operator))
        if self.max_docs:
            dbq.limit(self.max_docs)
        return dbq
    class Test:
        OPS = {
            "eq": "=", "=": "=",
            "ne": "<>", "<>": "<>", "!=": "<>",
            "lt": "<", "<": "<", "lte": "<=", "<=": "<=",
            "gt": ">", ">": ">", "gte": ">=", ">=": ">="
            }
        def __init__(self, assertion):
            try:
                path, operator, value = assertion.split(None, 2)
            except:
                raise ValueError("invalid test assertion %r" % assertion)
            assert path and value, "query test must have path and value"
            if path.startswith("/"):
                self.path, self.column = path.rsplit("/", 1)
                if self.column not in ("value", "int_val"):
                    raise ValueError("invalid query_term column %r" % self.path)
            elif path == "CdrCtl/Title":
                self.path, self.column = path, None
            else:
                raise ValueError("unsupported path %r" % test.path)
            if operator == "contains":
                self.value = "%{}%".format(value)
                self.operator = "LIKE"
            elif operator == "begins":
                self.value = "{}%".format(value)
                self.operator = "LIKE"
            else:
                self.value = value
                self.operator = self.OPS.get(operator)
                if not self.operator:
                    raise ValueError("unsupported operator %r" % operator)

if __name__ == "__main__":
    logging.basicConfig(level="DEBUG")
    parser = argparse.ArgumentParser()
    parser.add_argument("--doc-type", nargs="*")
    parser.add_argument("--test", nargs="*")
    parser.add_argument("--max_docs", type=int)
    args = parser.parse_args()
    result = Query(**vars(args)).run()
    print(str(etree.tostring(result, pretty_print=True), "utf-8"))
