import datetime
import logging
import re
import threading
from lxml import etree
from cdrlib import db

class Prototype:
    DIR = "C:/Users/bkline/Dropbox/cdr/new_server"
    LOG_DIR = "%s/logs" % DIR
    DATA_DIR = "%s/data" % DIR
    FILTER_DIR = "%s/filters/new" % DATA_DIR
    SCHEMA_DIR = "%s/schemas/new" % DATA_DIR
    VALIDATION_TEMPLATE = "%s/validation-template.xsd" % DATA_DIR

class Logging:
    FORMAT = "%(asctime)s [%(levelname)s] %(message)s"
    LEVELS = {
        "info": logging.INFO,
        "debug": logging.DEBUG,
        "warn": logging.WARN,
        "warning": logging.WARNING,
        "critical": logging.CRITICAL,
        "error": logging.ERROR
    }
    class Formatter(logging.Formatter):
        """Make our own logging formatter to get the time stamps right."""
        converter = datetime.datetime.fromtimestamp
        def formatTime(self, record, datefmt=None):
            ct = self.converter(record.created)
            if datefmt:
                s = ct.strftime(datefmt)
            else:
                t = ct.strftime("%Y-%m-%d %H:%M:%S")
                s = "%s.%03d" % (t, record.msecs)
            return s
    @classmethod
    def get_logger(cls, name, **opts):
        """
        Factory method for instantiating a logging object.

        name       required name for the logger
        path       optional path for the log file
        format     optional override for the default log format pattern
        level      optional verbosity for logging, defaults to info
        propagate  if True, the base handler also writes our entries
        multiplex  if True, add new handler even there already is one
        """

        logger = logging.getLogger(name)
        level = opts.get("level", "info")
        logger.setLevel(cls.LEVELS.get(level, logging.INFO))
        logger.propagate = opts.get("propagate", False)
        if not logger.handlers or opts.get("multiplex"):
            default_path = "%s/%s.log" % (Prototype.LOG_DIR, name)
            path = opts.get("path", default_path)
            handler = logging.FileHandler(path)
            formatter = cls.Formatter(opts.get("format", cls.FORMAT))
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        return logger

class Session:
    logger = logging.getLogger("cdr-server")
    def __init__(self, name=None):
        self.name = name
    def allowed(self, action, doctype=None):
        return True
    class Local(threading.local):
        def __init__(self, **kw):
            self.conn = db.connect()
            self.cursor = self.conn.cursor()
            self.docs = []
            self.__dict__.update(kw)
    local = Local()

class Doc:
    NS = "cips.nci.nih.gov/cdr"
    NSMAP = { "cdr": NS }
    def __init__(self, doc_id, doc_version=None):
        self.doc_id = doc_id
        self.doc_version = doc_version
    def int_id(self):
        return self.id_from_string(str(self.doc_id))
    @staticmethod
    def id_from_string(id):
        try:
            return int(re.sub(r"[^\d]+", "", str(id).split("#")[0]))
        except:
            Session.logger.warning("can't extract ID from %s", id)
            return None
    @staticmethod
    def doctype_from_id(doc_id):
        query = db.Query("doc_type t", "t.name")
        query.join("document d", "d.doc_type = t.id")
        query.where(query.Condition("d.id", doc_id))
        row = query.execute(Session.local.cursor).fetchone()
        return row and row[0] or None
    @staticmethod
    def id_from_title(title):
        title = title.replace("@@SLASH@@", "/").replace("+", " ")
        query = db.Query("document", "id")
        query.where(query.Condition("title", title))
        rows = query.execute(Session.local.cursor).fetchall()
        if len(rows) > 1:
            raise Exception("Multiple documents with title %s" % title)
        return rows and rows[0][0] or None
    @staticmethod
    def get_last_version_id(doc_id, publishable=False):
        try:
            doc_id = Doc.id_from_string(doc_id)
            query = db.Query("doc_version v", "MAX(v.num)")
            query.where(query.Condition("v.id", doc_id))
            if publishable:
                query.where("v.publishable = 'Y'")
            return query.execute(Session.local.cursor).fetchone()[0] or None
        except:
            return None
    @staticmethod
    def get_xml(doc_id, version=None):
        if version:
            query = db.Query("doc_version", "xml")
            query.where(query.Condition("num", version))
        else:
            query = db.Query("document", "xml")
        query.where(query.Condition("id", doc_id))
        row = query.execute(Session.local.cursor).fetchone()
        return row and row[0] or None
    @staticmethod
    def get_text(node, default=None):
        if node is None:
            return default
        return "".join(node.itertext("*"))

