import argparse
import datetime
import logging
import re
import sys
import unicodedata
from builtins import dict
from lxml import etree
from cdrlib import db, util, xslt

def validate_doc(**args):
    return Validator(**args).run()

class DTD:
    NS = "http://www.w3.org/2001/XMLSchema"
    SCHEMA = "{%s}schema" % NS
    ELEMENT = "{%s}element" % NS
    COMPLEX_TYPE = "{%s}complexType" % NS
    SIMPLE_TYPE = "{%s}simpleType" % NS
    SIMPLE_CONTENT = "{%s}simpleContent" % NS
    GROUP = "{%s}group" % NS
    INCLUDE = "{%s}include" % NS
    ATTRIBUTE = "{%s}attribute" % NS
    SEQUENCE = "{%s}sequence" % NS
    CHOICE = "{%s}choice" % NS
    RESTRICTION = "{%s}restriction" % NS
    EXTENSION = "{%s}extension" % NS
    ENUMERATION = "{%s}enumeration" % NS
    NAME_START_CATEGORIES = { "Ll", "Lu", "Lo", "Lt", "Nl" }
    OTHER_NAME_CATEGORIES = { "Mc", "Me", "Mn", "Lm", "Nd" }
    NAME_CHAR_CATEGORIES = NAME_START_CATEGORIES | OTHER_NAME_CATEGORIES
    EMPTY = "empty"
    MIXED = "mixed"
    TEXT_ONLY = "text-only"
    ELEMENT_ONLY = "element-only"
    UNBOUNDED = "unbounded"

    def __init__(self, **args):
        self.types = {}
        self.groups = {}
        self.top = None
        self.name = args.get("name")
        self.parse_schema(self.name)
    def parse_schema(self, name):
        util.Session.logger.debug("start parsing %s", name)
        root = Validator.get_doc(name)
        if root.tag != self.SCHEMA:
            raise Exception("Top-level element must be schema")
        for node in root:
            if node.tag == self.ELEMENT:
                assert not self.top, "only one top-level element allowed"
                self.top = self.Element(self, node)
            elif node.tag == self.COMPLEX_TYPE:
                self.ComplexType(self, node)
            elif node.tag == self.SIMPLE_TYPE:
                self.SimpleType(self, node)
            elif node.tag == self.GROUP:
                self.Group(self, node)
            elif node.tag == self.INCLUDE:
                self.parse_schema(node.get("schemaLocation"))
                util.Session.logger.debug("resume parsing %s", name)
        util.Session.logger.debug("finished parsing %s", name)
    def __str__(self):
        lines = ["<!-- Generated from %s -->" % self.name, ""]
        self.defined = set()
        self.top.define(lines)
        return "\n".join(lines) + "\n"
    def add_type(self, t):
        assert t not in self.types, "duplicate type %s" % t.name
        self.types[t.name] = t

    class CountedNode:
        def __init__(self, dtd, node):
            self.dtd = dtd
            self.min_occurs = self.max_occurs = 1
            min_occurs = node.get("minOccurs")
            max_occurs = node.get("maxOccurs")
            if min_occurs:
                self.min_occurs = int(min_occurs)
            if max_occurs == DTD.UNBOUNDED:
                self.max_occurs = sys.maxsize
            elif max_occurs:
                self.max_occurs = int(max_occurs)
            if self.min_occurs < 1:
                self.count_char = self.max_occurs == 1 and "?" or "*"
            else:
                self.count_char = self.max_occurs > 1 and "+" or ""
    class Element(CountedNode):
        def __init__(self, dtd, node):
            DTD.CountedNode.__init__(self, dtd, node)
            self.name = node.get("name")
            self.type_name = node.get("type")
            assert self.name, "element name required"
            assert self.type_name, "element type required"
            debug = util.Session.logger.debug
            debug("Element %s of type %s", self.name, self.type_name)
        def lookup_type(self):
            element_type = self.dtd.types.get(self.type_name)
            if element_type:
                return element_type
            if "xsd:" not in self.type_name:
                vals = self.name, self.type_name
                raise Exception("element %s: type %s not found" % vals)
            return DTD.Type(self.dtd, self.type_name)
        def define(self, lines):
            self.children = []
            lines += self.lookup_type().define(self)
            if not self.dtd.defined:
                self.add_control_definitions(lines)
            for child in self.children:
                if child.name not in self.dtd.defined:
                    self.dtd.defined.add(child.name)
                    child.define(lines)
        def add_control_definitions(self, lines):
            lines.append("<!ELEMENT CdrDocCtl (DocId, DocTitle)>")
            lines.append("<!ATTLIST CdrDocCtl readyForReview CDATA #IMPLIED>")
            lines.append("<!ELEMENT DocId (#PCDATA)>")
            lines.append("<!ATTLIST DocId readonly (yes|no) #IMPLIED>")
            lines.append("<!ELEMENT DocTitle (#PCDATA)>")
            lines.append("<!ATTLIST DocTitle readonly (yes|no) #IMPLIED>")
            for name in (self.name, "CdrDocCtl", "DocId", "DocTitle"):
                self.dtd.defined.add(name)
        def get_node(self, elements, serialize=False):
            if self.name not in [element.name for element in elements]:
                elements.append(self)
            if serialize:
                return "%s%s" % (self.name, self.count_char)
    class Type:
        def __init__(self, dtd, name):
            self.dtd = dtd
            self.name = name
            assert self.name, "type must have a name"
        def define(self, element):
            definitions = [self.define_element(element)]
            attributes = self.define_attributes(element)
            if attributes:
                definitions.append(attributes)
            return definitions
        def define_element(self, element):
            return "<!ELEMENT %s (#PCDATA)>" % element.name
        def define_attributes(self, element):
            return None
        @staticmethod
        def map_type(schema_type):
            for n in ("ID", "IDREFS", "NMTOKEN", "NMTOKENS"):
                if schema_type == "xsd:%s" % n:
                    return n
            return "CDATA"

    class SimpleType(Type):
        def __init__(self, dtd, node):
            DTD.Type.__init__(self, dtd, node.get("name"))
            self.base = None
            self.nmtokens = []
            util.Session.logger.debug("SimpleType %s", self.name)
            for restriction in node.findall(DTD.RESTRICTION):
                self.base = restriction.get("base")
                for enum in restriction.findall(DTD.ENUMERATION):
                    value = enum.get("value")
                    if DTD.is_nmtoken(value):
                        self.nmtokens.append(value)
                    else:
                        self.nmtokens = None
                        break
            dtd.add_type(self)
        def dtd_type(self):
            if self.nmtokens:
                return "(%s)" % "|".join(sorted(self.nmtokens))
            if "xsd:" in self.base:
                return DTD.Type.map_type(self.base)
            base = self.dtd.types.get(self.base)
            assert base, "%s: base type %s not found" % (self.name, self.base)
            return base.dtd_type()

    class ComplexType(Type):
        CONTENT = "sequence", "choice", "simpleContent", "group"
        CNAMES = ", ".join(["xsd:%s" % c for c in CONTENT])
        ERROR = "complex type may only contain one of %s" % CNAMES
        def __init__(self, dtd, node):
            DTD.Type.__init__(self, dtd, node.get("name"))
            util.Session.logger.debug("ComplexType %s", self.name)
            self.attributes = {}
            self.content = None
            self.model = DTD.EMPTY
            if node.get("mixed") == "true":
                self.model = DTD.MIXED
            for child in node.findall("*"):
                if child.tag == DTD.ATTRIBUTE:
                    self.add_attribute(dtd, child)
                elif self.content:
                    raise Exception("%s: %s" % (self.name, self.ERROR))
                elif child.tag == DTD.SIMPLE_CONTENT:
                    self.model = DTD.TEXT_ONLY
                    extension = child.find(DTD.EXTENSION)
                    assert len(extension), "%s: missing extension" % self.name
                    for child in extension.findall(DTD.ATTRIBUTE):
                        self.add_attribute(dtd, child)
                    break
                else:
                    if self.model == DTD.EMPTY:
                        self.model = DTD.ELEMENT_ONLY
                    if child.tag == DTD.SEQUENCE:
                        self.content = DTD.Sequence(dtd, child)
                    elif child.tag == DTD.CHOICE:
                        self.content = DTD.Choice(dtd, child)
                    elif child.tag == DTD.GROUP:
                        self.content = DTD.Group(dtd, child)
                    else:
                        raise Exception("%s: %s" % (self.name, self.ERROR))
            dtd.add_type(self)
        def add_attribute(self, dtd, child):
            attribute = DTD.Attribute(dtd, child)
            if attribute.name in self.attributes:
                values =  (self.name, attribute.name)
                error = "Duplicate attribute %s/@%s" % values
                raise Exception(error)
            self.attributes[attribute.name] = attribute
        def define_element(self, element):
            if self.model == DTD.TEXT_ONLY:
                return "<!ELEMENT %s (#PCDATA)>" % element.name
            elif self.model == DTD.EMPTY:
                return "<!ELEMENT %s EMPTY>" % element.name
            elif self.model == DTD.MIXED:
                self.content.get_node(element.children)
                children = "|".join([c.name for c in element.children])
                names = (element.name, children)
                return "<!ELEMENT %s (#PCDATA|%s)*>" % names
            elif self.model == DTD.ELEMENT_ONLY:
                content = self.content.get_node(element.children, True)
                assert content, "Elements required for elementOnly content"
                if not self.dtd.defined:
                    content = "(CdrDocCtl,%s)" % content
                elif not content.startswith("("):
                    content = "(%s)" % content
                return "<!ELEMENT %s %s>" % (element.name, content)
            raise Exception("%s: unrecognized content model" % self.name)
        def define_attributes(self, element):
            attributes = [str(a) for a in self.attributes.values()]
            if not self.dtd.defined and "readonly" not in self.attributes:
                attributes.append("readonly CDATA #IMPLIED")
            if attributes or not self.dtd.defined:
                attributes = sorted(attributes)
                if not self.dtd.defined:
                    attributes = ["xmlns:cdr CDATA #IMPLIED"] + attributes
                attributes = " ".join(attributes)
                return "<!ATTLIST %s %s>" % (element.name, attributes)

    class ChoiceOrSequence:
        def __init__(self, dtd, node):
            DTD.CountedNode.__init__(self, dtd, node)
            self.nodes = []
            for child in node:
                if child.tag == DTD.ELEMENT:
                    self.nodes.append(DTD.Element(dtd, child))
                if child.tag == DTD.CHOICE:
                    self.nodes.append(DTD.Choice(dtd, child))
                elif child.tag == DTD.SEQUENCE:
                    self.nodes.append(DTD.Sequence(dtd, child))
                elif child.tag == DTD.GROUP:
                    self.nodes.append(DTD.Group(dtd, child))
            assert self.nodes, "choice or sequence cannot be empty"
        def get_node(self, elements, serialize=False):
            nodes = [node.get_node(elements, serialize) for node in self.nodes]
            if serialize:
                string = self.separator.join(nodes)
                if len(self.nodes) > 1:
                    string = "(%s)" % string
                return "%s%s" % (string, self.count_char)
    class Choice(ChoiceOrSequence):
        separator = "|"
        def __init__(self, dtd, node):
            DTD.ChoiceOrSequence.__init__(self, dtd, node)
            util.Session.logger.debug("Choice with %d nodes", len(self.nodes))
    class Sequence(ChoiceOrSequence):
        separator = ","
        def __init__(self, dtd, node):
            DTD.ChoiceOrSequence.__init__(self, dtd, node)
            util.Session.logger.debug("Sequence with %d nodes", len(self.nodes))
    class Group:
        def __init__(self, dtd, node):
            self.dtd = dtd
            self.ref = node.get("ref")
            if self.ref:
                util.Session.logger.debug("Reference to group %s", self.ref)
                return
            self.name = node.get("name")
            util.Session.logger.debug("Group %s", self.name)
            if self.name in dtd.groups:
                raise Exception("multiple definitions for group %s" % self.name)
            nodes = []
            for child in node:
                if child.tag == DTD.CHOICE:
                    nodes.append(DTD.Choice(dtd, node))
                elif child.tag == DTD.SEQUENCE:
                    nodes.append(DTD.Sequence(dtd, node))
            assert len(nodes) == 1, "%s: %d nodes" % (self.name, len(nodes))
            self.node = nodes[0]
            dtd.groups[self.name] = self
        def get_node(self, elements, serialize=False):
            if self.ref:
                return self.dtd.groups[self.ref].get_node(elements, serialize)
            else:
                return self.node.get_node(elements, serialize)
    class Attribute:
        def __init__(self, dtd, node):
            self.dtd = dtd
            self.name = node.get("name")
            self.type_name = node.get("type")
            debug = util.Session.logger.debug
            debug("Attribute %s of type %s", self.name, self.type_name)
            self.required = node.get("use") == "required"
            if self.name.startswith("cdr-"):
                self.name = self.name.replace("cdr-", "cdr:")
        def __str__(self):
            required = self.required and "REQUIRED" or "IMPLIED"
            return "%s %s #%s" % (self.name, self.dtd_type(), required)
        def dtd_type(self):
            if "xsd:" in self.type_name:
                return DTD.Type.map_type(self.type_name)
            simple_type = self.dtd.types.get(self.type_name)
            if not simple_type:
                vals = self.type_name, self.name
                error = "unrecognized type %s for @%s" % vals
                raise Exception(error)
            return simple_type.dtd_type()
    @classmethod
    def alternate_is_nmtoken(cls, string):
        for c in string:
            o = ord(c)
            if 0x20DD < o <= 0x20E0:
                return False
            if 0xF900 < o < 0xFFFF:
                return False
            if c not in u".-_:\u00B7\u0e87":
                if unicodedata.category(c) not in cls.NAME_CHAR_CATEGORIES:
                    return False
        return True

    NMTOKEN_PATTERN = re.compile(
        r"[-:_."
        r"\xB7"
        r"0-9A-Za-z"
        r"\U000000C0-\U000000D6"
        r"\U000000D8-\U000000F6"
        r"\U000000F8-\U000002FF"
        r"\U00000300-\U0000036F"
        r"\U00000370-\U0000037D"
        r"\U0000037F-\U00001FFF"
        r"\U0000200C-\U0000200D"
        r"\U0000203F-\U00002040"
        r"\U00002070-\U0000218F"
        r"\U00002C00-\U00002FEF"
        r"\U00003001-\U0000D7FF"
        r"\U0000F900-\U0000FDCF"
        r"\U0000FDF0-\U0000FFFD"
        r"\U00010000-\U000EFFFF]+"
    )
    @classmethod
    def is_nmtoken(cls, string):
        """
        See https://www.w3.org/TR/REC-xml/#sec-common-syn
        """
        if " " in string:
            return False
        return cls.NMTOKEN_PATTERN.match(string) and True or False

class Validator:
    DEBUG = False
    logger = logging.getLogger("cdr-server")
    def __init__(self, **args):
        self.cursor = util.Session.local.cursor
        self.session = util.Session(args.get("session", "guest"))
        self.doc = None
        self.use_schema = self.val_links = True
        self.show_error_locators = self.update_db = False
        if "request" in args:
            self.parse_request(args["request"])
        else:
            for option in ("use_schema", "val_links", "update_db",
                           "show_error_locators"):
                if option in args:
                    setattr(self, option, args.get(option) and True or False)
            self.doc = self.Doc(self, **args)
        if not self.session.allowed("VALIDATE DOCUMENT", self.doc.type):
            raise Exception("Validation of %s documents is not authorized "
                            "for this user" % self.doc.type)
    def run(self):
        self.logger.info("validating CDR%s", self.doc.id)
        if self.doc.type not in self.Doc.CONTROL_TYPES:
            if self.use_schema:
                self.doc.validate_against_schema()
        response = etree.Element("CdrValidateDocResp")
        etree.SubElement(response, "DocId").text = self.doc.cdr_id
        etree.SubElement(response, "DocActiveStatus").text = self.doc.status
        if self.doc.errors:
            response.append(self.doc.errors_node())
            if self.show_error_locators:
                response.append(self.doc.cdr_doc_object())
        return response
    def parse_request(self, node):
        types = node.get("ValidationTypes")
        print("ValidationTypes=%s" % types)
        self.logger.debug("ValidationTypes: %s", types)
        if types:
            if "Schema" not in types:
                self.use_schema = False
            if "Links" not in types:
                self.val_links = False
        if node.get("ErrorLocators") == "Y":
            self.show_error_locators = True
        args = { "doctype": node.get("DocType") }
        for child in node:
            if child.tag == "DocId":
                self.doc = self.Doc(self, doc_id=child.text)
                if node.get("DocType") != self.doc.type:
                    raise Exception("Command specifies incorrect DocType")
                if child.get("ValidateOnly") != "Y":
                    self.update_db = True
            elif child.tag == "CdrDoc":
                self.doc = self.Doc(node=child)
    class Doc:
        CONTROL_TYPES = set(["Filter", "css", "schema"])
        NS = DTD.NS
        INCLUDE = DTD.INCLUDE
        ANNOTATION = "{%s}annotation" % NS
        APPINFO = "{%s}appinfo" % NS
        UNVALIDATED = "U"
        VALID = "V"
        INVALID = "I"
        MALFORMED = "M"
        VALIDATION = "validation"
        OTHER = "other"
        INFO = "info"
        WARNING = "warning"
        ERROR = "error"
        FATAL = "fatal"
        def __init__(self, validator, **args):
            self.validator = validator
            self.status = self.UNVALIDATED
            self.xml = self.root = None
            self.errors = []
            if "node" in args:
                self.parse_node(args["node"])
            else:
                self.id = util.Doc.id_from_string(args.get("doc_id"))
                self.type = args.get("doctype")
                self.xml = args.get("xml")
                self.root = args.get("root")
            if isinstance(self.xml, bytes):
                self.xml = str(self.xml, "utf-8")
            if not self.root:
                if not self.xml:
                    assert self.id, "nothing to validate"
                    self.xml = util.Doc.get_xml(self.id)
                    assert self.xml, "CDR%d not found" % self.id
                try:
                    self.root = etree.fromstring(bytes(self.xml, "utf-8"))
                except Exception as e:
                    Validator.logger.exception("parsing error; id=%s", self.id)
                    self.status = self.MALFORMED
                    self.add_error("document malformed", type=self.OTHER)
            if not self.type:
                assert self.id, "Validator.Doc(): need ID or doc type"
                self.type = util.Doc.doctype_from_id(self.id)
            if self.root is not None and not self.xml:
                self.xml = str(etree.tostring(self.root), "utf-8")
            if self.type in self.CONTROL_TYPES:
                self.status = self.VALID
            elif self.xml and re.search(r"[\uE000-\uF8FF]+", self.xml):
                self.add_error("Document contains private use character(s)")
            self.cdr_id = self.id and ("CDR%010d" % self.id) or ""
        def parse_node(self, node):
            self.type = node.get("Type")
            cdr_id = node.get("Id")
            self.id = cdr_id and util.Doc.id_from_string(cdr_id) or None
            self.xml = util.Doc.get_text(node.find("CdrDocXml"))
        def errors_node(self):
            errors = etree.Element("Errors", count=str(len(self.errors)))
            for error in self.errors:
                errors.append(error.to_node())
            return errors
        def add_error(self, message, location=None, **opts):
            for ncname in ("id", "href", "ref", "xref"):
                message = message.replace("cdr-%s" % ncname, "cdr:%s" % ncname)
            self.errors.append(self.Error(message, location, **opts))
        class Error:
            def __init__(self, message, location, **opts):
                self.message = re.sub(r"\s+", " ", message.strip())
                self.location = location
                self.type = opts.get("type") or Validator.Doc.VALIDATION
                self.level = opts.get("level") or Validator.Doc.ERROR
            def to_node(self):
                node = etree.Element("Err", etype=self.type, elevel=self.level)
                node.text = self.message
                if self.location:
                    node.set("eref", self.location)
                return node
        class LineMap:
            class Line:
                def __init__(self):
                    self.tags = {}
                    self.first = None
                def add_node(self, node):
                    if not self.first:
                        self.first = node.get("cdr-eid")
                    if not self.tags.get(node.tag):
                        self.tags[node.tag] = node.get("cdr-eid")
                def get_error_location(self, error):
                    match = re.match("Element '([^']+)'", error.message)
                    if not match:
                        Validator.logger.error("no element named in %r",
                                               error.message)
                        return None
                    tag = match.group(1) or None
                    location = self.tags.get(tag)
                    if not location:
                        Validator.logger.error("no node found for %r", tag)
                    return location or self.first
            def __init__(self, root):
                self.lines = {}
                for node in root.iter("*"):
                    line = self.lines.get(node.sourceline)
                    if not line:
                        line = self.lines[node.sourceline] = self.Line()
                    line.add_node(node)
            def get_error_location(self, error):
                line = self.lines.get(error.line)
                return line and line.get_error_location(error) or None
        def cdr_doc_object(self):
            if self.validator.use_schema:
                self.namespaces_on()
            doc_object = etree.Element("CdrDocXml")
            opts = { "pretty_print": True, "encoding": "utf-8" }
            doc_xml = etree.tostring(self.root, **opts)
            doc_object.text = etree.CDATA(doc_xml)
            return doc_object
        def validate_against_schema(self):
            if self.status == self.UNVALIDATED:
                self.status = self.VALID
            schema_doc = self.get_schema()
            schema = etree.XMLSchema(schema_doc)
            self.namespaces_off()
            self.strip_eids()
            if not schema.validate(self.root):
                self.status = self.INVALID
                location = None
                if self.validator.show_error_locators:
                    self.insert_eids()
                    line_map = self.LineMap(self.root)
                for error in schema.error_log:
                    if self.validator.show_error_locators:
                        location = line_map.get_error_location(error)
                    self.add_error(error.message, location)
            self.rule_sets = {}
            self.extract_rule_sets(schema_doc)
            for name in self.rule_sets:
                rule_sets = self.rule_sets[name]
                if not name:
                    name = "anonymous"
                Validator.logger.info("applying %s rule set(s)", name)
                opts = { "pretty_print": True, "encoding": "utf-8" }
                for rule_set in rule_sets:
                    filter_xml = etree.tostring(rule_set.get_xslt(), **opts)
                    if Validator.DEBUG == "capture validation filters":
                        open("val-filters.log", "ab").write(filter_xml + b"\n")
                    result = xslt.XSLT.apply_filter(filter_xml, self.root)
                    for node in result.doc.findall("Err"):
                        self.add_error(node.text, node.get("cdr-eid"))
                    #print(result.xml)
                    for entry in result.error_log:
                        self.status = self.INVALID
                        self.add_error(entry.message)
                        #print(entry.message)
        def namespaces_off(self):
            NS = "{%s}" % util.Doc.NS
            for node in self.root.iter("*"):
                for name in node.attrib:
                    if name.startswith(NS):
                        ncname = name.replace(NS, "cdr-")
                        node.set(ncname, node.get(name))
                        del node.attrib[name]
        def namespaces_on(self):
            NS = "{%s}" % util.Doc.NS
            for node in self.root.iter("*"):
                for name in node.attrib:
                    if name.startswith("cdr-"):
                        qname = name.replace("cdr-", NS)
                        node.set(qname, node.get(name))
                        del node.attrib[name]
        def strip_eids(self):
            for node in self.root.xpath("//*[@cdr-eid]"):
                del node.attrib["cdr-eid"]
        def insert_eids(self):
            eid = 1
            for node in self.root.iter("*"):
                node.set("cdr-eid", "_%d" % eid)
                eid += 1
        def extract_rule_sets(self, schema):
            for node in schema:
                if node.tag == self.ANNOTATION:
                    Validator.logger.debug("found xsd:annotation")
                    for child in node.findall("%s/pattern" % self.APPINFO):
                        Validator.logger.debug("found xsd:appinfo")
                        rule_set = Validator.RuleSet(child)
                        if rule_set.name not in self.rule_sets:
                            self.rule_sets[rule_set.name] = []
                        self.rule_sets[rule_set.name].append(rule_set)
                elif node.tag == self.INCLUDE:
                    name = node.get("schemaLocation")
                    self.extract_rule_sets(Validator.get_doc(name))
            path = "%s/%s/pattern" % (self.ANNOTATION, self.APPINFO)
        def get_schema(self):
            query = db.Query("document d", "d.title")
            query.join("doc_type t", "t.xml_schema = d.id")
            query.where(query.Condition("t.name", self.type))
            row = query.execute(self.validator.cursor).fetchone()
            assert row, "no schema for document type %s" % self.type
            return Validator.get_doc(row[0].replace(".xml", ".xsd"))
    class RuleSet:
        def __init__(self, node):
            self.name = node.get("name")
            self.value = node.get("value") # never been used, AFAIK
            self.rules = [Validator.Rule(r) for r in node.findall("rule")]
        def get_xslt(self):
            root = self.make_base()
            for rule in self.rules:
                Validator.logger.debug("context %s", rule.context)
                root.append(rule.template())
            return root
        @classmethod
        def make_base(cls):
            return etree.parse(util.Prototype.VALIDATION_TEMPLATE).getroot()
    class Rule:
        def __init__(self, node):
            self.assertions = []
            self.context = node.get("context")
            for child in node.findall("assert"):
                self.assertions.append(Validator.Assertion(child))
        def template(self):
            TEMPLATE = "{%s}template" % xslt.XSLT.NS
            template = etree.Element(TEMPLATE, match=self.context)
            for assertion in self.assertions:
                template.append(assertion.make_test())
            return template
    class Assertion:
        def __init__(self, node):
            self.test = util.Doc.get_text(node.find("test"))
            self.message = util.Doc.get_text(node.find("message"))
            assert self.test and self.message, "assert requires test, message"
        def make_test(self):
            IF = "{%s}if" % xslt.XSLT.NS
            CALL = "{%s}call-template" % xslt.XSLT.NS
            WITH = "{%s}with-param" % xslt.XSLT.NS
            test = etree.Element(IF, test="not(%s)" % self.test)
            call = etree.SubElement(test, CALL, name="pack-error")
            message = '"%s"' % self.message
            etree.SubElement(call, WITH, name="msg", select=message)
            return test
    @classmethod
    def get_doc(cls, name):
        class Resolver(etree.Resolver):
            def resolve(self, url, id, context):
                Validator.logger.debug("resolving %s", url)
                path = "%s/%s" % (util.Prototype.SCHEMA_DIR, url)
                return self.resolve_filename(path, context)
        cls.logger.info("fetching %s", name)
        assert name, "Validator.get_doc(): no name for schema"
        path = "%s/%s" % (util.Prototype.SCHEMA_DIR, name)
        try:
            parser = etree.XMLParser()
            parser.resolvers.add(Resolver())
            xml = open(path, "rb").read()
            return etree.fromstring(xml, parser)
        except Exception as e:
            cls.logger.exception("failure fetching schema")
            raise Exception("no valid schema found for %s" % name)

if __name__ == "__main__":
    Validator.DEBUG = True
    logging.basicConfig(level="DEBUG")
    parser = argparse.ArgumentParser()
    parser.add_argument("--doc-id")
    parser.add_argument("--use-schema", action="store_true")
    parser.add_argument("--show-error-locators", action="store_true")
    args = parser.parse_args()
    print(str(etree.tostring(Validator(**vars(args)).run(), pretty_print=True),
              "utf-8"))
