import argparse
import datetime
import logging
import re
import threading
import time
import urllib.parse
import lxml.etree as etree
from cdrlib import db

class Doc:
    NS = "cips.nci.nih.gov/cdr"
    NSMAP = { "cdr": NS }
    def __init__(self, doc_id, doc_version=None):
        self.doc_id = doc_id
        self.doc_version = doc_version
    def int_id(self):
        return self.id_from_string(str(self.doc_id))
    @staticmethod
    def id_from_string(id):
        return int(re.sub(r"[^\d]+", "", str(id).split("#")[0]))
    @staticmethod
    def doctype_from_id(doc_id):
        query = db.Query("doc_type t", "t.name")
        query.join("document d", "d.doc_type = t.id")
        query.where(query.Condition("d.id", doc_id))
        row = query.execute(Resolver.local.cursor).fetchone()
        return row and row[0] or None
    @staticmethod
    def id_from_title(title):
        title = title.replace("@@SLASH@@", "/").replace("+", " ")
        query = db.Query("document", "id")
        query.where(query.Condition("title", title))
        rows = query.execute(Resolver.local.cursor).fetchall()
        if len(rows) > 1:
            raise Exception("Multiple documents with title %s" % title)
        return rows and rows[0][0] or None
    @staticmethod
    def get_last_version_id(doc_id, publishable=False):
        try:
            doc_id = Doc.id_from_string(doc_id)
            query = db.Query("doc_version v", "MAX(v.num)")
            query.where(query.Condition("v.id", doc_id))
            if publishable:
                #query.join("active_doc a", "a.id = v.id")
                query.where("v.publishable = 'Y'")
            return query.execute(Resolver.local.cursor).fetchone()[0] or None
        except:
            return None
    @staticmethod
    def get_doc_xml(doc_id, version):
        if version:
            query = db.Query("doc_version", "xml")
            query.where(query.Condition("num", version))
        else:
            query = db.Query("document", "xml")
        query.where(query.Condition("id", doc_id))
        row = query.execute(Resolver.local.cursor).fetchone()
        return row and row[0] or None
    @staticmethod
    def get_text(node, default=None):
        if node is None:
            return default
        return "".join(node.itertext("*"))

class Term:
    MAX_DEPTH = 15
    MAX_CACHE_AGE = datetime.timedelta(1)
    use_cache = 0
    terms = {}
    lock = threading.Lock()
    cache_started = None
    def __init__(self, doc_id, depth=0):
        self.doc_id = doc_id
        self.cdr_id = "CDR%010d" % doc_id
        self.name = self.pdq_key = self.xml = self.full_xml = None
        self.include = True
        self.parents = {}
        self.version = Doc.get_last_version_id(doc_id, publishable=True)
        if self.version:
            xml = Doc.get_doc_xml(doc_id, self.version)
            root = etree.fromstring(bytes(xml, "utf-8"))
            self.name = Doc.get_text(root.find("PreferredName"))
            self.pdq_key = Doc.get_text(root.find("PdqKey"))
            for node in root.findall("TermType/TermTypeName"):
                if Doc.get_text(node) in ("Header term", "Obsolete term"):
                    self.include = False
                    break
            for node in root.findall("TermRelationship/ParentTerm/TermId"):
                self.get_parent(node, depth)
        if self.name and not depth:
            XSLT.logger.debug("serializing term with %d parents",
                              len(self.parents))
            self.serialize()
    def get_xml(self, with_upcoding=True):
        with Term.lock:
            if self.xml and self.full_xml:
                return with_upcoding and self.full_xml or self.xml
        self.serialize(need_locking=True)
        return with_upcoding and self.full_xml or self.xml
    def serialize(self, need_locking=False):
        term = etree.Element("Term", nsmap=Doc.NSMAP)
        term.set("{%s}ref" % Doc.NS, self.cdr_id)
        if self.pdq_key:
            term.set("PdqKey", "Term:%s" % self.pdq_key)
        etree.SubElement(term, "PreferredName").text = self.name
        xml = etree.tostring(term, encoding="utf-8")
        for doc_id in sorted(self.parents):
            parent = self.parents[doc_id]
            if parent is not None and parent.include and parent.name:
                child = etree.SubElement(term, "Term")
                child.set("{%s}ref" % Doc.NS, parent.cdr_id)
                if parent.pdq_key:
                    child.set("PdqKey", "Term:%s" % parent.pdq_key)
                etree.SubElement(child, "PreferredName").text = parent.name
        full_xml = etree.tostring(term, encoding="utf-8")
        if need_locking:
            with Term.lock:
                if not(self.xml and self.full_xml):
                    self.xml, self.full_xml = xml, full_xml
        else:
            self.xml, self.full_xml = xml, full_xml
    def get_parent(self, node, depth):
        try:
            doc_id = Doc.id_from_string(node.get("{%s}ref" % Doc.NS))
            XSLT.logger.debug("fetching term parent CDR%d", doc_id)
        except:
            failure = "No cdr:ref for parent of Term %s" % self.cdr_id
            raise Exception(failure)
        if doc_id not in self.parents:
            parent = Term.get_term(doc_id, depth + 1)
            if parent:
                self.parents.update(parent.parents)
                self.parents[doc_id] = parent
    @classmethod
    def get_term(cls, doc_id, depth=0):
        if depth > cls.MAX_DEPTH:
            raise Exception("term hierarchy depth exceeded at CDR%d" % doc_id)
        with cls.lock:
            if cls.use_cache:
                cache_age = datetime.datetime.now() - cls.cache_started
                if cache_age > cls.MAX_CACHE_AGE:
                    while cls.use_cache:
                        cls.set_cache(False)
            if doc_id in cls.terms:
                return cls.terms[doc_id]
        term = cls(doc_id, depth)
        if not term.name:
            term = None
        with cls.lock:
            if cls.use_cache and doc_id not in cls.terms:
                cls.terms[doc_id] = term
        return term
    @classmethod
    def set_cache(cls, on=True):
        with cls.lock:
            previous_state = cls.use_cache > 0
            if on:
                cls.use_cache += 1
                if not previous_state:
                    cls.cache_started = datetime.datetime.now()
            else:
                cls.use_cache -= 1
                if cls.use_cache < 0:
                    cls.use_cache = 0
                if not cls.use_cache:
                    cls.terms = {}
            return previous_state

class Local(threading.local):
    def __init__(self, **kw):
        self.conn = db.connect()
        self.cursor = self.conn.cursor()
        self.docs = []
        self.__dict__.update(kw)

class CdrUtil:
    @staticmethod
    def escape_uri(context, arg=""):
        if isinstance(arg, (list, tuple)):
            arg = "".join(arg)
        try:
            return urllib.parse.quote(arg.replace("+", "@@PLUS@@"))
        except:
            print("cdr:escape_uri(%r)" % arg)
            raise
    @staticmethod
    def date(context, arg=None):
        XSLT.logger.debug(context)
        XSLT.logger.debug(dir(context))
        XSLT.logger.debug(repr(context.context_node))
        XSLT.logger.debug(dir(context.context_node))
        #tree = context.context_node.getroottree()
        #parser = tree.parser
        #XSLT.logger.debug(parser.cdrdata)
        now = datetime.datetime.now()
        if arg:
            return now.strftime(arg)
        return str(now)[:19]

ns = etree.FunctionNamespace(Doc.NS)
ns["escape-uri"] = CdrUtil.escape_uri
#ns["valid-zip"] = CdrUtil.valid_zip
#ns["date"] = CdrUtil.date
#ns["sql-query"] = CdrUtil.date


class Resolver(etree.Resolver):
    UNSAFE = re.compile(r"insert\s|update\s|delete\s|create\s|alter\s"
                        r"exec[(\s]|execute[(\s]")
    ID_KEY_STRIP = re.compile("[^A-Z0-9]+")
    local = Local()
    #def __init__(self, prefix):
    #    self.prefix = prefix
    def resolve(self, url, pubid, context):
        self.url = urllib.parse.unquote(url.replace("+", " "))
        self.url = self.url.replace("@@PLUS@@", "+")
        if XSLT.DEBUG:
            fp = open("c:/tmp/filter-uri.log", "a")
            fp.write("%s\n" % self.url)
            fp.close()
        if url == "cdrx:/last":
            return self.resolve_string("<empty/>", context)
        self.cursor = self.local.cursor
        XSLT.logger.debug("resolving %r", self.url)
        scheme, parms = self.url.split(":", 1)
        parms = parms.strip("/")
        if scheme in ("cdr", "cdrx"):
            return self.get_doc(parms, context)
        elif scheme == "cdrutil":
            return self.run_function(parms, context)
        raise Exception("unsupported url %r" % self.url)
    def run_function(self, parms, context):
        function, args = parms, None
        if "/" in parms:
            function, args = parms.split("/", 1)
        if function == "docid":
            return self.get_doc_id(context)
        elif function == "sql-query":
            return self.run_sql_query(args, context)
        elif function == "get-pv-num":
            return self.get_pv_num(args, context)
        elif function == "denormalizeTerm":
            return self.get_term(args, context)
        elif function == "dedup-ids":
            return self.dedup_ids(args, context)
        elif function == "valid-zip":
            return self.valid_zip(args, context)
        XSLT.logger.debug("function", function)
        raise Exception("unsupported function %r in %r" % (function, self.url))
    @classmethod
    def make_id_key(cls, id):
        return cls.ID_KEY_STRIP.sub("", id.upper())
    def valid_zip(self, args, context):
        result = etree.Element("ValidZip")
        query = db.Query("zipcode", "zip")
        query.where(query.Condition("zip", args))
        row = query.execute(self.cursor).fetchone()
        if row and row[0]:
            result.text = str(row[0])[:5]
        return self.package_result(result, context)
    def dedup_ids(self, args, context):
        ids = []
        skip = set()
        if "~~" in args:
            primary, secondary = [i.split("~") for i in args.split("~~", 1)]
            for p in primary:
                skip.add(self.make_id_key(p))
            for s in secondary:
                key = self.make_id_key(s)
                if key and key not in skip:
                    ids.append(s)
                    skip.add(key)
        result = etree.Element("result")
        for i in ids:
            etree.SubElement(result, "id").text = i
        return self.package_result(result, context)
    def package_result(self, result, context):
        result = etree.tostring(result, encoding="utf-8")
        return self.resolve_string(result, context)
    def get_term(self, args, context):
        if "/" in args:
            doc_id = Doc.id_from_string(args.split("/")[0])
            upcode = False
        else:
            doc_id = Doc.id_from_string(args)
            upcode = True
        term = Term.get_term(doc_id)
        if term is None:
            XSLT.logger.warning("term CDR%s not available", doc_id)
            term_xml = "<empty/>"
        else:
            term_xml = term.get_xml(upcode)
            if XSLT.DEBUG == "show upcoding":
                suffix = upcode and "upcoded" or "plain"
                name = "term-%s-%s.xml" % (doc_id, suffix)
                open(name, "w").write(repr(term_xml))
        return self.resolve_string(term_xml, context)
    def get_pv_num(self, args, context):
        ver = Doc.get_last_version_id(args, publishable=True) or 0
        answer = etree.Element("PubVerNumber")
        answer.text = str(ver)
        return self.package_result(answer, context)
    def run_sql_query(self, args, context):
        if "~" in args:
            query, values = args.split("~", 1)
            values = values.split("~")
        else:
            query, values = args, []
        if self.UNSAFE.search(query):
            raise Exception("query contains disallowed sql keywords")
        if query.count("?") != len(values):
            raise Exception("wrong number of sql query placeholder values")
        cursor = self.local.cursor
        if db.Query.PLACEHOLDER != "?":
            query = query.replace("?", db.Query.PLACEHOLDER)
        try:
            cursor.execute(query, tuple(values))
        except Exception as e:
            XSLT.logger.exception("%s\n%r", query, values)
            raise
        names = [col[0] for col in cursor.description]
        result = etree.Element("SqlResult")
        r = 1
        for values in cursor.fetchall():
            row = etree.SubElement(result, "row", id=str(r))
            for c, v in enumerate(values):
                col = etree.SubElement(row, "col", id=str(c), name=names[c])
                if v is None:
                    col.set("null", "Y")
                else:
                    col.text = str(v)
            r += 1
        return self.package_result(result, context)
    def get_doc_id(self, context):
        doc_id = self.local.docs[-1].doc_id
        element = etree.Element("DocId")
        element.text = doc_id
        return self.package_result(element, context)
    def get_doc(self, parms, context):
        if parms.startswith("*"):
            return self.get_metadata(parms, context)
        if parms.startswith("name:"):
            parms = parms[5:]
            if "/" in parms:
                title, version = parms.split("/", 1)
            else:
                title, version = parms, None
            doc_id = Doc.id_from_title(title)
            if not doc_id:
                XSLT.logger.warning("no document for title %r", title)
                return None
            if Doc.doctype_from_id(doc_id) == "Filter":
                return self.get_filter(doc_id, context)
            parms = str(doc_id)
            if version:
                parms = "%d/%s" % (doc_id, version)
        return self.get_doc_xml(parms, context)
    def get_doc_xml(self, parms, context):
        doc, version = parms, None
        if "/" in parms:
            doc, version = parms.split("/", 1)
        if not doc:
            raise Exception("no document specified")
        try:
            doc_id = Doc.id_from_string(doc)
        except:
            XSLT.logger.warning("no document found for %r", self.url)
            return self.resolve_string("", context)
        version = self.get_version_number(doc_id, version)
        doc_xml = Doc.get_doc_xml(doc_id, version)
        if not doc_xml:
            XSLT.logger.error("can't find doc xml for %r", parms)
            return None
        if XSLT.DEBUG:
            path = "documents/%s-%s.xml" % (doc_id, version or "cwd")
            open(path, "wb").write(bytes(doc_xml, "utf-8"))
        return self.resolve_string(doc_xml, context)
    def get_metadata(self, parms, context):
        doc_id = self.local.docs[-1].int_id()
        if "/CdrCtl" in parms:
            return self.get_doc_ctl(doc_id, context)
        elif "/DocTitle" in parms:
            return self.get_doc_title(doc_id, context)
        else:
            raise Exception("unsupported url %r" % self.url)
    def get_filter(self, doc_id, context):
        doc_xml = XSLT.get_filter(doc_id, self.cursor).xml
        return self.resolve_string(doc_xml, context)
    def get_version_number(self, doc_id, version):
        if not version:
            return None
        if version in ("last", "lastp"):
            version = Doc.get_last_version_id(doc_id, version=="lastp")
            if not version:
                raise Exception("version not found for %d" % doc_id)
            return version
        try:
            return int(version)
        except:
            raise Exception("unrecognized version %s for CDR%s in %s" %
                            (version, doc_id, self.url))
    def get_doc_title(self, doc_id, context, version=None):
        if version:
            query = db.Query("doc_version", "title")
            query.where(query.Condition("num", version))
        else:
            query = db.Query("document", "title")
        query.where(query.Condition("id", doc_id))
        row = query.execute(self.cursor).fetchone()
        title = row and row[0] or ""
        element = etree.Element("CdrDocTitle")
        element.text = title
        return self.package_result(element, context)

    def get_doc_ctl(self, doc_id, context):
        cols = "d.val_status", "d.val_date", "d.title", "d.comment", "r.doc_id"
        query = db.Query("document d", *cols)
        query.where(query.Condition("d.id", doc_id))
        query.outer("ready_for_review r", "r.doc_id = d.id")
        row = query.execute(self.cursor).fetchone()
        status, date, title, comment, rr_id = row
        element = etree.Element("CdrDocCtl")
        etree.SubElement(element, "DocValStatus").text = status
        if date:
            date = str(date).replace(" ", "T")
            etree.SubElement(element, "DocValDate").text = date
        etree.SubElement(element, "DocTitle").text = title
        if comment:
            etree.SubElement(element, "DocComment").text = comment
        rr = rr_id and "Y" or "N"
        etree.SubElement(element, "ReadyForReview").text = rr
        query = db.Query("audit_trail t", "t.dt", "u.name")
        query.join("usr u", "u.id = t.usr")
        query.join("action a", "a.id = t.action")
        query.where(query.Condition("t.document", doc_id))
        query.where("a.name = 'ADD DOCUMENT'")
        dt, name = query.execute(self.cursor).fetchone()
        create = etree.SubElement(element, "Create")
        etree.SubElement(create, "Date").text = str(dt).replace(" ", "T")
        etree.SubElement(create, "User").text = name
        query = db.Query("audit_trail t", "t.dt", "u.name")
        query.join("usr u", "u.id = t.usr")
        query.join("action a", "a.id = t.action")
        query.where(query.Condition("t.document", doc_id))
        query.where("a.name = 'MODIFY DOCUMENT'")
        query.order("t.dt DESC").limit(1)
        row = query.execute(self.cursor).fetchone()
        if row:
            date, user = row
            date = str(date).replace(" ", "T")
            modify = etree.SubElement(element, "Modify")
            etree.SubElement(modify, "Date").text = date
            etree.SubElement(modify, "User").text = user
        date = self.get_date_first_pub(doc_id)
        if date:
            first = etree.SubElement(element, "FirstPub")
            etree.SubElement(first, "Date").text = date.replace(" ", "T")
        return self.package_result(element, context)
    @classmethod
    def get_date_first_pub(cls, doc_id):
        cursor = cls.local.cursor
        query = db.Query("document", "first_pub")
        query.where("first_pub IS NOT NULL")
        query.where("first_pub_knowable = 'Y'")
        query.where(query.Condition("id", doc_id))
        row = query.execute(cursor).fetchone()
        return row and str(row[0]) or None #and str(row[0])[:10] or None

class Parser(etree.XMLParser):
    def __init__(self):
        etree.XMLParser.__init__(self)
        self.resolvers.add(Resolver("cdrutil"))
        self.resolvers.add(Resolver("cdr"))
        self.resolvers.add(Resolver("cdrx"))

class XSLT:
    DEBUG = False
    MAX_FILTER_SET_DEPTH = 20
    NS = "http://www.w3.org/1999/XSL/Transform"
    filter_set_lock = threading.Lock()
    filter_lock = threading.Lock()
    filters = {}
    filter_sets = {}
    logger = logging.getLogger("cdr-server")
    num_filters = None
    def __init__(self, request):
        self.cursor = Resolver.local.cursor
        self.output = request.get("Output") != "N"
        self.filters = []
        self.parms = {}
        names = "doc", "doc_id", "max_doc_date", "max_filter_date", "doc_ver"
        for name in names:
            setattr(self, name, None)
        for node in request:
            if node.tag == "Document":
                self.doc_id = node.get("href")
                self.doc_ver = node.get("version")
                self.doc = self.get_doc(self.doc_id, self.doc_ver)
            elif node.tag == "FilterSet":
                set_name = node.get("Name")
                self.filters += self.get_filter_set(set_name)
            elif node.tag == "Parms":
                for child in node.findall("Parm"):
                    name = child.find("Name").text
                    value = child.find("Value").text
                    self.parms[name] = etree.XSLT.strparam(value)
                    #print("set %s=%s" % (name, value))
                    #self.parms[name] = value
        XSLT.logger.info("filtering %s-V%s with %s (parms=%s)", self.doc_id,
                         self.doc_ver, set_name, self.parms)
    class FilterResult:
        def __init__(self, doc, error_log):
            self.doc = doc
            self.error_log = error_log
    @classmethod
    def apply_filter(cls, filter_xml, doc, parser=None, **parms):
        if not parser:
            parser = Parser()
        transform = etree.XSLT(etree.fromstring(filter_xml, parser))
        doc = transform(doc, **parms)
        return cls.FilterResult(doc, transform.error_log)
    def filter_doc(self):
        Resolver.local.docs.append(Doc(self.doc_id, self.doc_ver))
        response = None
        try:
            parser = Parser()
            doc = self.doc
            messages = []
            if XSLT.num_filters is not None:
                filters = self.filters[:XSLT.num_filters]
            else:
                filters = self.filters
            if not filters:
                raise Exception("no filters specified")
            for f in filters:
                #print("filtering with CDR%010d" % f.doc_id)
                XSLT.logger.debug("filtering with CDR%010d", f.doc_id)
                result = self.apply_filter(f.xml, doc, parser, **self.parms)
                doc = result.doc
                for entry in result.error_log:
                    messages.append(entry.message)
            response = etree.Element("CdrFilterResp")
            opts = {
                "xml_declaration": True,
                "encoding": "utf-8"
            }
            if XSLT.DEBUG:
                opts["pretty_print"] = True
            suffix = "xml"
            output = filter_doc.find("{%s}output" % self.NS)
            if output is not None:
                if output.get("method") == "html":
                    opts["method"] = "html"
                    opts["doctype"] = "<!DOCTYPE html>"
                    suffix = "html"
                if output.get("encoding"):
                    opts["encoding"] = output.get("encoding")
                if output.get("indent") == "yes":
                    opts["pretty_print"] = True
            document = etree.tostring(doc, **opts)
            if XSLT.DEBUG:
                open("output.%s" % suffix, "wb").write(document)
            document = str(document, "utf-8")
            etree.SubElement(response, "Document").text = etree.CDATA(document)
            if messages:
                wrapper = etree.SubElement(response, "Messages")
                for message in messages:
                    XSLT.logger.warning(message)
                    etree.SubElement(wrapper, "message").text = message
        finally:
            Resolver.local.docs.pop()
        return response
    def get_filter_set(self, name):
        query = db.Query("filter_set", "id")
        query.where(query.Condition("name", name))
        row = query.execute(self.cursor).fetchone()
        if not row:
            return []
        return self.get_filter_set_by_id(row[0])
    def get_filter_set_by_id(self, set_id, depth=0):
        if depth > self.MAX_FILTER_SET_DEPTH:
            raise Exception("infinite filter set recursion")
        depth += 1
        with XSLT.filter_set_lock:
            if set_id in XSLT.filter_sets:
                return XSLT.filter_sets[set_id]
        fields = "filter", "subset", "position"
        query = db.Query("filter_set_member", *fields).order("position")
        query.where(query.Condition("filter_set", set_id))
        filters = []
        for filter_id, subset_id, pos in query.execute(self.cursor).fetchall():
            if filter_id:
                filters.append(self.get_filter(filter_id, self.cursor))
            elif subset_id:
                filters += self.get_filter_set_by_id(subset_id, depth)
        with XSLT.filter_set_lock:
            if set_id not in XSLT.filter_sets:
                XSLT.filter_sets[set_id] = filters
        return filters
    #@staticmethod
    #def get_title(doc_xml):
    #    doc_xml = str(doc_xml, "utf-8")
    #    match = re.search(r"<!--\s*Filter title:\s*(.*)\s*-->", doc_xml)
    #    if not match:
    #        print(doc_xml)
    #        raise Exception("unable to find title")
    #        return "unable to find title"
    #    return match.group(1)
    @classmethod
    def get_filter(cls, doc_id, cursor):
        with cls.filter_lock:
            f = cls.filters.get(doc_id)
            if f is not None:
                if f.stale():
                    del cls.filters[doc_id]
                else:
                    return f
        #query = db.Query("doc_version", "xml", "num")
        #query.order("num DESC").limit(1)
        query = db.Query("good_filters", "xml")
        query.where(query.Condition("id", doc_id))
        #xml, ver = query.execute(cursor).fetchone()
        xml = query.execute(cursor).fetchone()[0]
        root = etree.fromstring(bytes(xml, "utf-8"))
        for name in ("import", "include"):
            for node in root.findall("{%s}%s" % (cls.NS, name)):
                href = node.get("href")
                node.set("href", href.replace(" ", "%20"))
        xml = etree.tostring(root.getroottree(), encoding="utf-8")
        with cls.filter_lock:
            if doc_id not in cls.filters:
                if XSLT.DEBUG:
                    open("filters/CDR%010d.xml" % doc_id, "wb").write(xml)
                cls.filters[doc_id] = cls.Filter(doc_id, xml)
        return cls.filters[doc_id]
    def get_doc(self, doc_id, doc_ver):
        doc_id = Doc.id_from_string(doc_id)
        xml = bytes(Doc.get_doc_xml(doc_id, doc_ver), "utf-8")
        if XSLT.DEBUG:
            name = "%s-%s.xml" % (doc_id, doc_ver or "cwd")
            open("documents/%s" % name, "wb").write(xml)
        return etree.fromstring(xml)
    class Filter:
        def __init__(self, doc_id, xml):
            self.doc_id, self.xml = doc_id, xml
            self.now = time.time()
        def stale(self):
            now = time.time()
            answer = (now - self.now) > 60
            self.now = now
            return answer
    @classmethod
    def test(cls):
        cls.DEBUG = True
        Term.set_cache()
        logging.basicConfig(level="DEBUG")
        parser = argparse.ArgumentParser()
        parser.add_argument("--id", "-i", required=True, type=int)
        parser.add_argument("--version", "-v")
        parser.add_argument("--filter-set", "-s", required=True)
        parser.add_argument("--num-filters", "-n", type=int)
        args = parser.parse_args()
        request = etree.Element("CdrFilter")
        document = etree.SubElement(request, "Document")
        document.set("href", "CDR%010d" % args.id)
        if args.version:
            document.set("version", args.version)
        if args.num_filters:
            XSLT.num_filters = args.num_filters
        filter_set = etree.SubElement(request, "FilterSet")
        filter_set.set("Name", args.filter_set)
        date = Resolver.get_date_first_pub(args.id)
        if date:
            print("DateFirstPub=%s" % date)
            parms = etree.SubElement(request, "Parms")
            parm = etree.SubElement(parms, "Parm")
            etree.SubElement(parm, "Name").text = "DateFirstPub"
            etree.SubElement(parm, "Value").text = date
        xslt = cls(request)
        opts = { "pretty_print": True, "encoding": "utf-8" }
        xml = etree.tostring(xslt.filter_doc(), **opts)
        print(str(xml, "ascii", "ignore"))
        Term.set_cache(False)

if __name__ == "__main__":
    XSLT.test()
