"""
Here are the numbers from publishing filtering using the C++ server:
2017-03-26 16:54:38,068:INFO:elapsed time for CTGovProtocol: 547.692
2017-03-26 16:55:14,811:INFO:elapsed time for DrugInformationSummary: 36.741
2017-03-26 17:10:51,546:INFO:elapsed time for GlossaryTermName: 936.736
2017-03-26 17:42:25,713:INFO:elapsed time for Organization: 1894.165
2017-03-26 17:45:58,413:INFO:elapsed time for Person: 212.698
2017-03-26 17:46:13,197:INFO:elapsed time for PoliticalSubUnit: 14.783
2017-03-26 18:34:06,122:INFO:elapsed time for Summary: 2872.924
2017-03-26 18:52:10,730:INFO:elapsed time for Term: 1084.605
"""

import argparse
import datetime
import logging
import socket
import struct
import threading
import lxml.etree as etree

def send_command(command):
    command_set = etree.Element("CdrCommandSet")
    etree.SubElement(command_set, "SessionId").text = Doc.session
    wrapper = etree.SubElement(command_set, "CdrCommand")
    wrapper.append(command)
    request = etree.tostring(command_set, encoding="utf-8")
    length = len(request)
    length_bytes = struct.pack("!L", length)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(("127.0.0.1", 2019))
    try:
        sock.send(length_bytes)
        sock.sendall(request)
        length_bytes = sock.recv(4)
        if len(length_bytes) != 4:
            raise Exception("server did not send response length")
        length, = struct.unpack("!L", length_bytes)
        if not length:
            raise Exception("server did not send response")
        received = 0
        chunks = []
        while received < length:
            response_bytes = sock.recv(8192)
            if not response_bytes:
                break
            chunks.append(response_bytes)
            received += len(response_bytes)
        if received < length:
            raise Exception("Expected {} bytes reading server response; "
                            "got {} bytes".format(length, received))
        return b"".join(chunks)
    finally:
        sock.close()

class Doc:
    docs = []
    next_doc = 0
    next_doc_lock = threading.Lock()
    done = ("CTGovProtocol", "DrugInformationSummary",
            "GlossaryTermName", "Organization", "Person")
    done = ()
    aliases = {
        "Person": "GeneticsProfessional",
        "GlossaryTermName": "GlossaryTerm",
        "DrugInformationSummary": "DrugInfoSummary"
    }
    def __init__(self, doc_id, doc_version, doc_type, first_pub):
        self.doc_id = doc_id
        self.doc_version = doc_version
        self.doc_type = doc_type
        self.first_pub = first_pub
    def get_vendor_filter_set(self):
        return "Vendor %s Set" % self.aliases.get(self.doc_type, self.doc_type)
    def run_filter_set(self):
        cdr_id = "CDR%010d" % self.doc_id
        fset = self.get_vendor_filter_set()
        logging.info("thread %s:filtering %s %sV%s using %s",
                     threading.current_thread().ident,
                     self.doc_type, cdr_id, self.doc_version, fset)
        command = etree.Element("CdrFilter")
        document = etree.SubElement(command, "Document")
        document.set("href", cdr_id)
        document.set("version", str(self.doc_version))
        filter_set = etree.SubElement(command, "FilterSet")
        filter_set.set("Name", self.get_vendor_filter_set())
        if self.first_pub:
            wrapper = etree.SubElement(command, "Parms")
            parm = etree.SubElement(wrapper, "Parm")
            etree.SubElement(parm, "Name").text = "DateFirstPub"
            etree.SubElement(parm, "Value").text = self.first_pub
        response = send_command(command)
        root = etree.fromstring(response)
        for node in root.findall("CdrResponse/CdrFilterResp"):
            for message in node.findall("Messages/message"):
                logging.warning("%s: %s", cdr_id, message.text)
        #print(root.tag)
        doc = root.find("CdrResponse/CdrFilterResp/Document")
        #print("elapsed: %s" % root.find("CdrResponse").get("Elapsed"))
        if doc is not None:
            fp = open("new-pub/CDR%010d.xml" % self.doc_id, "wb")
            fp.write(bytes(doc.text, "utf-8"))
            fp.close()
        else:
            self.log_error(cdr_id, root, response)
    def log_error(self, cdr_id, root, response):
        errors = root.findall("CdrResponse/Errors/Err")
        if errors:
            for error in errors:
                logging.error("%s: %s", cdr_id, error.text)
        else:
            logging.error("%s: %r", cdr_id, response)

class Thread(threading.Thread):
    def run(self):
        while True:
            with Doc.next_doc_lock:
                if Doc.next_doc >= len(Doc.docs):
                    break
                doc = Doc.docs[Doc.next_doc]
                Doc.next_doc += 1
            doc.run_filter_set()

def main():
    logging.basicConfig(
        level=logging.DEBUG,
        filename="new-filter-test.log",
        format="%(asctime)s:%(levelname)s:%(message)s",
    )
    parser = argparse.ArgumentParser()
    parser.add_argument("--per-doctype", type=int, default=1)
    parser.add_argument("--num-threads", type=int, default=2)
    parser.add_argument("--skip", type=int, default=0)
    parser.add_argument("--session", default="guest")
    parser.add_argument("--doctype")
    args = parser.parse_args()
    Doc.session = args.session
    rows = eval(open("pub_docs.txt").read())
    logging.info("%d docs publishable", len(rows))
    doc_types = {}
    for doc_id, doc_version, doc_type, first_pub, first_pub_knowable in rows:
        if doc_type not in doc_types:
            doc_types[doc_type] = []
        if first_pub_knowable != "Y":
            first_pub = None
        else:
            first_pub = first_pub and str(first_pub)[:10] or None
        doc = Doc(doc_id, doc_version, doc_type, first_pub)
        doc_types[doc_type].append(doc)
    send_command(etree.Element("CdrCacheing", on="term"))
    logging.info("started filtering")
    start = datetime.datetime.now()
    processed = 0
    for doc_type in sorted(doc_types):
        if doc_type in Doc.done:
            continue
        if args.doctype and doc_type != args.doctype:
            continue
        docs = doc_types[doc_type][args.skip:]
        Doc.docs = docs[:args.per_doctype]
        logging.info("queued %d of %d %s documents", len(Doc.docs),
                     len(doc_types[doc_type]), doc_type)
        Doc.next_doc = 0
        threads = []
        doctype_start = datetime.datetime.now()
        for _ in range(args.num_threads):
            threads.append(Thread())
        for t in threads:
            t.start()
            #logging.debug("started thread %s", t.ident)
        for t in threads:
            t.join()
        elapsed = (datetime.datetime.now() - doctype_start).total_seconds()
        logging.info("elapsed time for %s: %s", doc_type, elapsed)
        processed += len(Doc.docs)
    elapsed = (datetime.datetime.now() - start).total_seconds()
    logging.info("total processing time: %s", elapsed)
    logging.info("total number of documents: %s", processed)
    logging.info("seconds per document: %s", elapsed / processed)
    logging.info("documents per second: %s", processed / elapsed)
    send_command(etree.Element("CdrCacheing", off="term"))
main()
