import argparse
import datetime
import socket
import struct
import lxml.etree as etree
import cdrdb3 as cdrdb

def send_command(command):
    command_set = etree.Element("CdrCommandSet")
    etree.SubElement(command_set, "SessionId").text = "guest"
    wrapper = etree.SubElement(command_set, "CdrCommand")
    wrapper.append(command)
    request = etree.tostring(command_set, encoding="utf-8")
    length = len(request)
    length_bytes = struct.pack("!L", length)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(("127.0.0.1", 2019))
    try:
        sock.send(length_bytes)
        sock.sendall(request)
        length_bytes = sock.recv(4)
        if len(length_bytes) != 4:
            raise Exception("server did not send response length")
        length, = struct.unpack("!L", length_bytes)
        if not length:
            raise Exception("server did not send response")
        received = 0
        chunks = []
        while received < length:
            response_bytes = sock.recv(8192)
            if not response_bytes:
                break
            chunks.append(response_bytes)
            received += len(response_bytes)
        if received < length:
            raise Exception("Expected {} bytes reading server response; "
                            "got {} bytes".format(length, received))
        return b"".join(chunks)
    finally:
        sock.close()

class Doc:
    aliases = {
        "Person": "GeneticsProfessional",
        "GlossaryTermName": "GlossaryTerm",
        "DrugInformationSummary": "DrugInfoSummary"
    }
    def __cmp__(self, other):
        return cmp(self.doc_id, other.doc_id)
    def __init__(self, doc_id, doc_version, doc_type, first_pub):
        self.doc_id = doc_id
        self.doc_version = doc_version
        self.doc_type = doc_type
        self.first_pub = first_pub
    def get_vendor_filter_set(self):
        return "Vendor %s Set" % self.aliases.get(self.doc_type, self.doc_type)
    def run_filter_set(self):
        command = etree.Element("CdrFilter")
        document = etree.SubElement(command, "Document")
        document.set("href", "CDR%010d" % self.doc_id)
        document.set("version", str(self.doc_version))
        filter_set = etree.SubElement(command, "FilterSet")
        filter_set.set("Name", self.get_vendor_filter_set())
        if self.first_pub:
            wrapper = etree.SubElement(command, "Parms")
            parm = etree.SubElement(wrapper, "Parm")
            etree.SubElement(parm, "Name").text = "DateFirstPub"
            etree.SubElement(parm, "Value").text = self.first_pub
        response = send_command(command)
        root = etree.fromstring(response)
        print(root.tag)
        doc = root.find("CdrResponse/CdrFilterResp/Document")
        print("elapsed: %s" % root.find("CdrResponse").get("Elapsed"))
        if doc is not None:
            fp = open("new-pub/CDR%010d.xml" % self.doc_id, "wb")
            fp.write(bytes(doc.text, "utf-8"))
            fp.close()
        else:
            print(str(response, "utf-8"))
            raise Exception("failure filtering CDR%d" % self.doc_id)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--per-doctype", type=int, default=1)
    args = parser.parse_args()
    cursor = cdrdb.connect("CdrGuest").cursor()
    fields = ("p.doc_id", "p.doc_version", "t.name", "d.first_pub",
              "d.first_pub_knowable")
    query = cdrdb.Query("pub_proc_doc p", *fields)
    query.join("pub_proc_cg c", "c.id = p.doc_id", "c.pub_proc = p.pub_proc")
    query.join("document d", "d.id = c.id")
    query.join("doc_type t", "t.id = d.doc_type")
    query.where("t.name <> 'Media'")
    query.order("p.doc_id")
    print(query)
    rows = query.execute(cursor).fetchall()
    print("%d rows fetched" % len(rows))
    open("pub_docs.txt", "w").write(repr(rows))
    doc_types = {}
    for doc_id, doc_version, doc_type, first_pub, first_pub_knowable in rows:
        if doc_type not in doc_types:
            doc_types[doc_type] = []
        if first_pub_knowable != "Y":
            first_pub = None
        else:
            first_pub = first_pub and str(first_pub)[:10] or None
        doc = Doc(doc_id, doc_version, doc_type, first_pub)
        doc_types[doc_type].append(doc)
    print(str(send_command(etree.Element("CdrCacheing", on="term"))))
    start = datetime.datetime.now()
    print("started filtering at %s" % start)
    for doc_type in sorted(doc_types):
        if doc_type != "Person":
            continue
        substart = datetime.datetime.now()
        for doc in doc_types[doc_type][:args.per_doctype]:
            doc.run_filter_set()
        print("%s %s" % ((datetime.datetime.now() - substart).total_seconds(),
                         doc_type))
    print("%s TOTAL" % (datetime.datetime.now() - start).total_seconds())
    print(str(send_command(etree.Element("CdrCacheing", off="term"))))
main()
