import argparse
import datetime
import logging
import socket
import struct
import threading
import lxml.etree as etree

def send_command(command):
    command_set = etree.Element("CdrCommandSet")
    etree.SubElement(command_set, "SessionId").text = Doc.session
    wrapper = etree.SubElement(command_set, "CdrCommand")
    wrapper.append(command)
    request = etree.tostring(command_set, encoding="utf-8")
    length = len(request)
    length_bytes = struct.pack("!L", length)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(("127.0.0.1", 2019))
    try:
        sock.send(length_bytes)
        sock.sendall(request)
        length_bytes = sock.recv(4)
        if len(length_bytes) != 4:
            raise Exception("server did not send response length")
        length, = struct.unpack("!L", length_bytes)
        if not length:
            raise Exception("server did not send response")
        received = 0
        chunks = []
        while received < length:
            response_bytes = sock.recv(8192)
            if not response_bytes:
                break
            chunks.append(response_bytes)
            received += len(response_bytes)
        if received < length:
            raise Exception("Expected {} bytes reading server response; "
                            "got {} bytes".format(length, received))
        return b"".join(chunks)
    finally:
        sock.close()

class Doc:
    qc_parms = {
        "DisplayCitations": "Y",
        "DisplayComments": "E",
        "DisplayGlossaryTermList": "Y",
        "DisplayImages": "N",
        "DisplayLOETermList": "N",
        "DurationComments": "A",
        "IncludeExtPerm": "N",
        "IncludeIntAdv": "Y",
        "SourceComments": "A",
        "displayBoard": "editorial-board_",
        "insRevLevels": "approved|",
    }
    queue = []
    next_doc = 0
    next_doc_lock = threading.Lock()
    def __init__(self, doc_id, doc_version):
        self.doc_id = doc_id
        self.doc_version = doc_version
    def run_filter_set(self):
        cdr_id = "CDR%010d" % self.doc_id
        logging.info("thread %s:filtering %sV%s using %s",
                     threading.current_thread().ident,
                     cdr_id, self.doc_version, Doc.filter_set)
        command = etree.Element("CdrFilter")
        document = etree.SubElement(command, "Document")
        document.set("href", cdr_id)
        document.set("version", str(self.doc_version))
        filter_set = etree.SubElement(command, "FilterSet")
        filter_set.set("Name", Doc.filter_set)
        for name in sorted(Doc.qc_parms):
            wrapper = etree.SubElement(command, "Parms")
            parm = etree.SubElement(wrapper, "Parm")
            etree.SubElement(parm, "Name").text = name
            etree.SubElement(parm, "Value").text = Doc.qc_parms[name]
        response = send_command(command)
        root = etree.fromstring(response)
        for node in root.findall("CdrResponse/CdrFilterResp"):
            for message in node.findall("Messages/message"):
                logging.warning("%s: %s", cdr_id, message.text)
        doc = root.find("CdrResponse/CdrFilterResp/Document")
        if doc is not None:
            fp = open("new-qc/CDR%010d.html" % self.doc_id, "wb")
            fp.write(bytes(doc.text.strip(), "utf-8"))
            fp.write(b"\n")
            fp.close()
        else:
            self.log_error(cdr_id, root, response)
    def log_error(self, cdr_id, root, response):
        errors = root.findall("CdrResponse/Errors/Err")
        if errors:
            for error in errors:
                logging.error("%s: %s", cdr_id, error.text)
        else:
            logging.error("%s: %r", cdr_id, response)

class Thread(threading.Thread):
    def run(self):
        while True:
            with Doc.next_doc_lock:
                if Doc.next_doc >= len(Doc.queue):
                    break
                doc = Doc.queue[Doc.next_doc]
                Doc.next_doc += 1
            doc.run_filter_set()

def main():
    logging.basicConfig(
        level=logging.DEBUG,
        filename="new-qc-test.log",
        format="%(asctime)s:%(levelname)s:%(message)s",
    )
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-docs", type=int, default=1)
    parser.add_argument("--num-threads", type=int, default=2)
    parser.add_argument("--skip", type=int, default=0)
    parser.add_argument("--session", default="guest")
    parser.add_argument("--filter-set", default="QC Summary Set")
    args = parser.parse_args()
    Doc.session = args.session
    Doc.filter_set = args.filter_set
    rows = eval(open("hp-summaries.txt").read())
    docs = []
    for doc_id, doc_version in rows:
        docs.append(Doc(doc_id, doc_version))
    send_command(etree.Element("CdrCacheing", on="term"))
    logging.info("started filtering")
    Doc.queue = docs[args.skip:][:args.num_docs]
    logging.info("queued %d of %d Summaries", len(Doc.queue), len(docs))
    Doc.next_doc = 0
    threads = []
    start = datetime.datetime.now()
    for _ in range(args.num_threads):
        threads.append(Thread())
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    elapsed = (datetime.datetime.now() - start).total_seconds()
    logging.info("elapsed time: %s", elapsed)
    processed = len(Doc.queue)
    elapsed = (datetime.datetime.now() - start).total_seconds()
    logging.info("total processing time: %s", elapsed)
    logging.info("total number of documents: %s", processed)
    logging.info("seconds per document: %s", elapsed / processed)
    logging.info("documents per second: %s", processed / elapsed)
    send_command(etree.Element("CdrCacheing", off="term"))
main()
